%******************************************************************************;
%* SAS macro MetaData vs. 0.0.4 , 29 Jan 2014, by Jim Groeneveld, Netherlands *;
%******************************************************************************;
%* STILL UNDER CONSTRUCTION, NOT FOR VALIDATION, USE WITH CAUTION AT OWN RISK *;
%*                                                                            *;
   %MACRO   MetaData   /* overview of metadata of many datasets, libraries    */
/* -------------------------------------------------------------------------- */
/* ---------- A l w a y s   s p e c i f i e d   a r g u m e n t s ----------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
(  LibList = WORK      /* Library list, default just WORK                   R */
,  DataList= _ALL_     /* dataset name list or _ALL_ or * for all datasets  R */
,  MetaData= MetaData  /* MetaData output dataset spec, dafault MetaData    R */
,  OverWrit= No        /* Yes: overwrite existing MetaData output dataset   R */
/*                                                                            */
/* ---------- R a r e l y   s p e c i f i e d   a r g u m e n t s ----------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  Version = 0.0.4     /* forced version specification for version control  O */
,  CalledBy=           /* Macro name of macro or program for err0r report   O */
,  Presence=           /* if not empty value is returned checking existence O */
/*                                                                            */
/* ----------- N e v e r   s p e c i f i e d   a r g u m e n t s ------------ */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  Debug   = 0         /* Debug level; for testing, debugging purposes only O */
,  ForMyUse=           /* argument for the program's use only, don't use it O */
)/ DES     =              'overview of metadata of many datasets'  /*Store*/   ;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* File       : MetaData.SAS / MetaData.004                                   *;
%*                                                                            *;
%* Purpose    : Overview of metadata (variables) of many datasets, libraries. *;
%* Features   : Already support for SAS name literal variable names.          *;
%*                                                                            *;
%* Input      : LibName (list), Dataset (list) or _ALL_ specification         *;
%*                                                                            *;
%* Output     : Overview of metadata (variables) of multiple datasets.        *;
%*                                                                            *;
%* Programmer : Jim Groeneveld                                                *;
%* Date       : 29 Jan 2014                                                   *;
%* Version    : 0.0.4                                                         *;
%*                                                                            *;
%* SAS version: 9.3 under Windows 7                                           *;
%*                                                                            *;
%* Validation : -                                                             *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Download location                                                          *;
%* -----------------                                                          *;
%* http://jim.groeneveld.eu.tf/software/SASmacro/MetaData.zip                 *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Example macro call                                                         *;
%* ------------------                                                         *;
/* %MetaData (LibList=WORK SomeLib, DataList=_ALL_,                           */
/*            MetaData=LibName.MetaData, Overwrit=Yes);                       */
%*                                                                            *;
%* Checking presence of macro in calling code:                                *;
%* - - - - - - - - - - - - - - - - - - - - - -                                *;
/* %LET Checking = %MetaData(Presence=any text);                              */
/* %IF (&Checking NE any text) %THEN ......... (macro not present) .........; */
%* or:                                                                        *;
/* %LET Checking = %MetaData(Presence=%NRSTR(&MacName._&MacVs);               */
/* %IF (&Checking NE MetaData_0.0.4) %THEN ..... (MacroVs not present) .....; */
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Used auxiliary macros:                                                     *;
%* ----------------------                                                     *;
%* = internal: -                                                              *;
%* = attached: -                                                              *;
/* = external: %UstrMcmp, %ERR0R, %WARN1NG, %_Delete_                         */
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* History of MetaData:                                                       *;
%* ====================                                                       *;
%* version  date         description of changes, additions, improvements      *;
%* -------  -----------  ---------------------------------------------------- *;
%* 0.0.0    19 Oct 2011  first Q(uick)&D(irty) working release (prototype),   *;
%*                         allowing a single library name as well as a single *;
%*                         dataset name or the keyword _ALL_ (all datasets).  *;
%* 0.0.1    20 Oct 2011  added defaults for DataLib, Data and MetaData        *;
%*                         if specified empty.                                *;
%* 0.0.2    06 Sep 2013  added library and dataset list specifications,       *;
%*    _BACKWARDS_          while changing DataLib argument into LibList and   *;
%*   _INCOMPATIBLE_        Data into DataList, added Overwrit argument,       *;
%*                         making this version backwards incompatible with    *;
%*                         the previous (not yet distributed) ones (OK),      *;
%* 0.0.3    13 Sep 2013  fixed slight bug with excluding PROC CONTENTS output,*;
%*                       added Presence argument checking macro`s existence,  *;
%*                       added full standard macro header layout (this one),  *;
%*                         including various arguments checks (consistency)   *;
%*                       added cleaning up code at the end (delete, reset),   *;
%*                       added adapting SYSLAST automatic macro variable      *;
%* 0.0.4    29 Jan 2014  fixed bug by declaring macro variable I LOCAL        *;
%*                                                                            *;
%* Future of MetaData:                                                        *;
%* ===================                                                        *;
%* - dataset options support (already experimentally programmed separately)   *;
%*                                                                            *;
%******************************************************************************;

  %LOCAL I   /* Define internal macro variables explicitely local */
    MacName  /* This macros external name: MetaData */
    MacVs    /* Version number major.minor.patchSub of macro: 0.0.4 */
    MacDate  /* Date of this version of macro: 29 Jan 2014 */
    ErrCount /* number of erroneously specified arguments */
    ArgList  /* List of user specifiable macro arguments */
    Argument /* Single user specifiable macro argument */
    Library  /* element of LibList */
    LibraryN /* number of library element in Liblist */
    Dataset  /* element of DataList */
    DatasetN /* number of dataset element in DataList */
    MetaLib  /* library specification of MetaData (to avoid refecencing) */
    MetaMem  /* dataset specification of MetaData (to avoid refecencing) */
    MetaWork /* dataset specification of MetaWork (to avoid referencing),
                intermediate output dataset of PROC CONTENTS */
    ValidVar /* stored VALIDVARNAME option */
    NoMPrint /* stored MPRINT option */
    NoMErr0r /* stored MERR0R option */
    NoSErr0r /* stored SERR0R option */
    NoMLogic /* stored MLOGIC option */
    Symbolgn /* stored SYMBOLGEN option */
    MacroGen /* stored MACROGEN option */
    NoSpool  /* stored SPOOL option */
  ;

  %LET MacName  = MetaData;
  %LET MacVs    = 0.0.4 ;
  %LET MacDate  = 29 Jan 2014;
  %LET ValidVar = %SYSFUNC(GETOPTION(VALIDVARNAME));
  %LET NoMPrint = %SYSFUNC(GETOPTION(MPRINT));
  %LET NoMErr0r = %SYSFUNC(GETOPTION(MERROR));
  %LET NoSErr0r = %SYSFUNC(GETOPTION(SERROR));
  %LET NoMLogic = %SYSFUNC(GETOPTION(MLOGIC));
  %LET Symbolgn = %SYSFUNC(GETOPTION(SYMBOLGEN));
  %LET MacroGen = %SYSFUNC(GETOPTION(MACROGEN));
  %LET NoSpool  = %SYSFUNC(GETOPTION(SPOOL));
  %LET MetaWork = __Cntnts; %* Contents;

  %IF (%LENGTH(%SUPERQ(Presence))) %THEN %DO;
    %UNQUOTE(&Presence) /* no semicolon, value returned */
    %GOTO Finish;
  %END;

  %IF ( %UPCASE(&ForMyUse) = COMPILE ) %THEN %GOTO Finish; %* Compilation;

  %IF (&Debug EQ 1 OR &Debug GT 11) %THEN %DO;
    OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN SPOOL;
  %END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Macro header in LOG file;
%*_________________________;

  %PUT;
  %PUT |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|;
  %PUT |   Macro &MacName, vs. &MacVs, by Jim Groeneveld, &MacDate  |;
  %PUT | Overview of metadata (variables) of many datasets, libraries |;
  %PUT |______________________________________________________________|;
  %PUT;

%* Feedback of specified arguments, programmed as macro loop;
  %PUT &MacName: Review of argument specifications (incl. defaults);
  %LET ArgList = LibList DataList MetaData Overwrit Version CalledBy Debug;
  %LET I = 1;
  %LET Argument = %SCAN ( &ArgList, &I, %STR( ) );
  %DO %WHILE (%LENGTH(&Argument));
    %PUT &MacName: &Argument = &&&Argument;
    %LET I = %EVAL ( &I + 1 );
    %LET Argument = %SCAN ( &ArgList, &I, %STR( ) );
  %END;

  %IF (&Version NE AND &Version NE &MacVs) %THEN
  %DO;
    %PUT *** &MacName *** %ERR0R-: Specified version &Version does not match &MacVs,;
    %PUT .                       macro &MacName will abort;
    %GOTO Exit;
  %END;

%* Argument checks;
%* LibList argument;
  %IF (NOT %LENGTH(%SUPERQ(LibList))) %THEN %DO;
    %PUT *** &MacName *** %WARN1NG-: LibList specified empty but is required,;
    %PUT .                     LibList set to default WORK;
    %LET LibList = WORK;
  %END;

%* DataList argument;
  %IF (NOT %LENGTH(%SUPERQ(DataList))) %THEN %DO;
    %PUT *** &MacName *** %WARN1NG-: DataList specified empty but is required,;
    %PUT .                     DataList set to default _ALL_;
    %LET DataList = _ALL_;
  %END;
  %ELSE %IF (%SUPERQ(DataSet) EQ %STR(*)) %THEN %LET DataSet = _ALL_; %* *-spec.;

%* MetaData argument;
  %IF (NOT %LENGTH(%SUPERQ(MetaData))) %THEN %DO;
    %PUT *** &MacName *** %WARN1NG-: MetaData specified empty but is required,;
    %PUT .                     MetaData set to default MetaData (in WORK);
    %LET MetaData = MetaData;
  %END;

  %IF ( NOT %UstrMcmp(&Overwrit,Yes) AND NOT %UstrMcmp(&Overwrit,No) ) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Overwrit argument not specified or illegal: &Overwrit,;
    %PUT .                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1 );
  %END;

  %IF ( %SYSFUNC(EXIST(&MetaData)) AND NOT %UstrMcmp(&OverWrit,Yes) ) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Dataset MetaData=&MetaData exists and OverWrit is not Yes,;
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1 );
  %END;

  %IF ( &ErrCount GT 0 ) %THEN %GOTO Aborting;

%* - - - - - - - - - - - ;
%* Core code starts here ;
%* - - - - - - - - - - - ;

%* Determine library and dataset name of MetaData;
  %IF (%LENGTH(%SCAN(&MetaData, 2, .))) %THEN %DO; * library specification;
    %LET MetaLib = %SCAN(&MetaData, 1, .);
    %LET MetaMem = %SCAN(&MetaData, 2, .);
  %END;
  %ELSE %DO; %* no library specification;
    %LET MetaLib = WORK;
    %LET MetaMem = &MetaData;
  %END;

**OPTIONS VALIDVARNAME=ANY;

* remove existing, old result dataset (Overwrit=Yes);
  %IF ( %SYSFUNC(EXIST(&MetaLib..&MetaMem)) ) %THEN %DO;
    PROC DATASETS LIBRARY=&MetaLib; DELETE &MetaMem; RUN;
  %END;
* empty result dataset as a start (not with PROC APPEND) ;
/*DATA &MetaData; STOP; RUN;*/

  %LET LibraryN = 1;
  %LET Library = %SCAN ( &LibList, &LibraryN, %STR( ) );
  %DO %WHILE (%LENGTH(%SUPERQ(Library)));
    %LET DatasetN = 1;
    %LET Dataset = %SCAN ( &DataList, &DatasetN, %STR( ) );
    %DO %WHILE (%LENGTH(%SUPERQ(Dataset)));
%* Generate Meta information from PROC CONTENTS for ;
%* each existing Library and Dataset combination.   ;
      %IF (%UPCASE(&Dataset) EQ _ALL_ OR
           %SYSFUNC(EXIST(&Library..&Dataset))) %THEN %DO;
        PROC DATASETS LIB=&Library MEMTYPE=DATA NOLIST;
          CONTENTS DATA=&Dataset NOPRINT
            OUT=WORK.&MetaWork (KEEP = LibName MemName Name Label Type
                                       Format FormatL FormatD Length
                          /* prevent MetaData dataset itself in overview */
                          WHERE = (NOT (UPCASE(MemName) EQ UPCASE("&MetaMem")
                                    AND UPCASE(LibName) EQ UPCASE("&MetaLib") )
                                    OR (UPCASE(MemName) EQ UPCASE("&MetaWork")
                                    AND UPCASE(LibName) EQ "WORK") ) );
        RUN; QUIT;
* Append &MetaWork to &MetaData, a fixed structure;
        PROC APPEND BASE=&MetaData DATA=&MetaWork; RUN;
      %END; %* Existing dataset;
      %LET DatasetN = %EVAL ( &DatasetN + 1 );
      %LET Dataset = %SCAN ( &DataList, &DatasetN, %STR( ) );
    %END;
    %LET LibraryN = %EVAL ( &LibraryN + 1 );
    %LET Library = %SCAN ( &LibList, &LibraryN, %STR( ) );
  %END;

  %LET SYSLAST = &MetaData;

%Aborting:
%* - - - - - - ;
%* Cleaning up ;
%* - - - - - - ;

%* Restore possibly modified option settings to previous ones;
  %if (&Debug) %then %do;
    OPTIONS &NoMPrint &NoMErr0r &NoSErr0r &NoMLogic &Symbolgn &MacroGen &NoSpool;
  %end;
  OPTIONS VALIDVARNAME = &ValidVar;

%* Delete work datasets;
  %_Delete_ (&MetaWork);

%Exit:
%Finish:

%MEND MetaData; %* - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
